/*
 * bnpairings/src/eta_tate.c version 20080222
 * Peter Schwabe
 * Public domain
*/

#include "fp.h"
#include "fp2e.h"
#include "fp6e.h"
#include "fp12e.h"
#include "curvepoint_fp.h"
#include "twistpoint_fp2.h"
#include "init.h"
#include "points.h"
#include "final_expo.h"
#include "eta_tate.h"

#include <stdio.h>
#include <gmp.h>
#include <cpucycles.h>

static void linefunction_evaluate_eta(fp12e_t rop, const fpe_t slope_numerator, const fpe_t slope_denominator, const fpe_t zapow3, const curvepoint_fp_t op1, const twistpoint_fp2_t op2)
{
	// Compute the Miller element (see Implementing Cryptographic Pairings over BN-Curves, page 5):

	fpe_t tfpe1, tfpe2;			// Temporary variables
	fp2e_t tfp2e1, tfp2e2, tfp2e3, tfp2e4;	// Temporary variables
	fp6e_t tfp6e1, tfp6e2;			// Temporary variables

	// Compute y_Q' * z_A^3 * z_C (Compare "Implementing Cryptographic Pairings over Barreto Naehrig Curves"):
	fpe_mul(tfpe1, zapow3, slope_denominator);

	fp2e_mul_fpe(tfp2e1, op2->m_y, tfpe1);
	fp2e_neg(tfp2e1, tfp2e1);

	// Compute x_Q' * z_A^3 * \lambda_A,B:
	fpe_mul(tfpe1, zapow3, slope_numerator);
	fp2e_mul_fpe(tfp2e2, op2->m_x, tfpe1);

	// Compute x_A * z_A * \lambda_A,B - y_A * z_C:
	fpe_mul(tfpe1, op1->m_x, op1->m_z);
	fpe_mul(tfpe1, tfpe1, slope_numerator);
	fpe_mul(tfpe2, op1->m_y, slope_denominator);
	fpe_sub(tfpe1, tfpe2, tfpe1);

	fp2e_set_fpe(tfp2e3, tfpe1);

	// Dummy set to 0 in F_p^2:
	fp2e_setzero(tfp2e4);

	fp6e_set_fp2e(tfp6e1, tfp2e4, tfp2e1, tfp2e4);
	fp6e_set_fp2e(tfp6e2, tfp2e4, tfp2e2, tfp2e3);
	fp12e_set_fp6e(rop, tfp6e1, tfp6e2);
}


static void linefunction_add_eta(fp12e_t rop1, curvepoint_fp_t rop2, const curvepoint_fp_t op1, const curvepoint_fp_t op2, const twistpoint_fp2_t op3)
{
	// op2 is assumed to be in affine coordinates!

	fpe_t tfpe1, tfpe2, tfpe3, tfpe4, tfpe5;	// Temporary variables

	// Computation of the slope (numerator and denominator), see Guide to ECC, page 91f.
	fpe_square(tfpe2, op1->m_z);
	fpe_mul(tfpe3, tfpe2, op1->m_z);

	// Store z_1^3 for use in the evaluation:
	fpe_set(tfpe4, tfpe3);	

	fpe_mul(tfpe2, tfpe2, op2->m_x);
	fpe_mul(tfpe3, tfpe3, op2->m_y);
	fpe_sub(tfpe2, tfpe2, op1->m_x);
	fpe_sub(tfpe3, tfpe3, op1->m_y);
	fpe_mul(tfpe1, tfpe2, op1->m_z); // z-coordinate computed, stored in tfpe0

	// Compute the linefunction:
	linefunction_evaluate_eta(rop1, tfpe3, tfpe1, tfpe4, op1, op3);

	// Copy z-coordinate to rop:
	fpe_set(rop2->m_z, tfpe1);

	// Continue computation of point coordinates:
	fpe_square(tfpe4, tfpe2);
	fpe_mul(tfpe5, tfpe4, tfpe2);
	fpe_mul(tfpe4, tfpe4, op1->m_x);
	fpe_double(tfpe2, tfpe4);
	fpe_square(rop2->m_x, tfpe3);
	fpe_sub(rop2->m_x, rop2->m_x, tfpe2);
	fpe_sub(rop2->m_x, rop2->m_x, tfpe5);
	fpe_sub(tfpe4, tfpe4, rop2->m_x);
	fpe_mul(tfpe4, tfpe4, tfpe3);
	fpe_mul(tfpe5, tfpe5, op1->m_y);
	fpe_sub(rop2->m_y, tfpe4, tfpe5);
}

static void linefunction_double_eta(fp12e_t rop1, curvepoint_fp_t rop2, const curvepoint_fp_t op1, const twistpoint_fp2_t op3)
{
	fpe_t tfpe1, tfpe2, tfpe3, tfpe4, tfpe5;	// Temporary variables

	// Computation of the slope (numerator and denominator), see DA Peter Schwabe, page 42
	fpe_square(tfpe2, op1->m_y);
	fpe_mul(tfpe3, op1->m_x, tfpe2);
	fpe_double(tfpe3, tfpe3);
	fpe_double(tfpe3, tfpe3);

	fpe_square(tfpe4, tfpe2);
	fpe_double(tfpe4, tfpe4);
	fpe_double(tfpe4, tfpe4);
	fpe_double(tfpe4, tfpe4);

	fpe_square(tfpe2, op1->m_x);
	fpe_triple(tfpe2, tfpe2);

	// Compute z_1^3 for evaluation
	fpe_square(tfpe5, op1->m_z);
	fpe_mul(tfpe5, tfpe5, op1->m_z);

	fpe_mul(tfpe1, op1->m_y, op1->m_z);
	fpe_double(tfpe1, tfpe1);

	// Compute the linefunction:
	linefunction_evaluate_eta(rop1, tfpe2, tfpe1, tfpe5, op1, op3);

	// Copy z-coordinate
	fpe_set(rop2->m_z, tfpe1);

	fpe_square(rop2->m_x, tfpe2);
	fpe_double(tfpe1, tfpe3);
	fpe_sub(rop2->m_x, rop2->m_x, tfpe1);
	fpe_sub(rop2->m_y, tfpe3, rop2->m_x);
	fpe_mul(rop2->m_y, rop2->m_y, tfpe2);
	fpe_sub(rop2->m_y, rop2->m_y, tfpe4);
}


void eta(fp12e_t rop, const curvepoint_fp_t op1, const twistpoint_fp2_t op2)
{
	fp12e_setone(rop);

	fp12e_t dummy;

	curvepoint_fp_t r;
	curvepoint_fp_set(r, op1);

	size_t i;

	for(i = mpz_sizeinbase(looplength_eta, 2) - 1; i > 0; i--)
	{
		linefunction_double_eta(dummy, r, r, op2);
		
		fp12e_square(rop, rop);
		fp12e_mul(rop, rop, dummy);

		if (mpz_tstbit(looplength_eta, i - 1)) 
		{
			linefunction_add_eta(dummy, r, r, op1, op2);
			fp12e_mul(rop, rop, dummy);
		}
	}
	
	final_expo(rop);
		
}

void tate(fp12e_t rop, const curvepoint_fp_t op1, const twistpoint_fp2_t op2)
{
	fp12e_setone(rop);

	fp12e_t dummy;

	curvepoint_fp_t r;
	curvepoint_fp_set(r, op1);

	size_t i;

	for(i = mpz_sizeinbase(n, 2) - 1; i > 0; i--)
	{
		linefunction_double_eta(dummy, r, r, op2);
		fp12e_square(rop, rop);
		fp12e_mul(rop, rop, dummy);

		if (mpz_tstbit(n, i - 1)) 
		{
			linefunction_add_eta(dummy, r, r, op1, op2);
			fp12e_mul(rop, rop, dummy);
		}
	}

	final_expo(rop);

}

